#!/usr/bin/python
from __future__ import print_function
import mysql.connector
from mysql.connector import errorcode
import json
from decimal import Decimal
from boConfig import DB_USER, DB_PASS, DB_HOST, DB_PORT, DB_NAME, LOGLEVEL
import logging
import sys
from geographiclib.geodesic import Geodesic

MAX_STORABLE_NUMBER_OF_CHANNELS = 8 # `channels` is an unsigend tinyint which can hold values in the range of 0..255 or with other words 8 bits
MAX_STORABLE_NUMBER_OF_CHANNELS_MASK = (2 ** MAX_STORABLE_NUMBER_OF_CHANNELS) - 1

TABLES = {}
TABLES['signals'] = (
    "CREATE TABLE `signals` ("
    "  `id` int(11) UNSIGNED NOT NULL AUTO_INCREMENT,"
    "  `time_ns` bigint(11) UNSIGNED NOT NULL,"
    "  `channels` tinyint(4) UNSIGNED NOT NULL,"
    "  PRIMARY KEY (`id`),"
    "  UNIQUE KEY `id` (`id`),"
    "  UNIQUE KEY `time_ns` (`time_ns`),"
    "  INDEX(`channels`)"
    ") ENGINE=InnoDB DEFAULT CHARSET=utf8;"
    )

TABLES['strokes'] = (
    "CREATE TABLE `strokes` ("
    "  `id` int(11) UNSIGNED NOT NULL AUTO_INCREMENT,"
    "  `signal_id` int(11) NULL DEFAULT NULL,"
    "  `time_ns_stroke` bigint(11) UNSIGNED NOT NULL,"
    "  `lat_stroke` decimal(9,6) NOT NULL,"
    "  `lon_stroke` decimal(9,6) NOT NULL,"
    "  `delta_time_ns_station` INT(11) NOT NULL,"
    "  `lat_station` decimal(9,6) NOT NULL,"
    "  `lon_station` decimal(9,6) NOT NULL,"
    "  `distance` decimal(6,1) NOT NULL,"
    "  `azimuth` decimal(4,1) NOT NULL,"
    "  `channels` tinyint(4) UNSIGNED NOT NULL,"
    "  PRIMARY KEY (`id`),"
    "  UNIQUE KEY `id` (`id`),"
    "  UNIQUE KEY `signal_id` (`signal_id`),"
    "  UNIQUE KEY `time_ns_stroke_delta_station` (`time_ns_stroke`,`delta_time_ns_station`),"
    "  INDEX(`azimuth`),"
    "  INDEX(`channels`),"
    "  INDEX(`distance`),"
    "  INDEX(`time_ns_stroke`)"
    ") ENGINE=InnoDB DEFAULT CHARSET=utf8;"
    )

TABLES['historicalData'] = (
    "CREATE TABLE `historicalData` ("
    "  `id` INT NOT NULL AUTO_INCREMENT, "
    "  `entity` TINYTEXT NOT NULL,"
    "  `attribute` TINYTEXT NOT NULL,"
    "  `value` TEXT NOT NULL,"
    "  PRIMARY KEY (`id`)"
    ") ENGINE=InnoDB DEFAULT CHARSET=utf8;"
    )

cnx = None
cursor = None

def connectMysql():
    global cnx
    global cursor
    cnx = mysql.connector.connect(user = DB_USER, password = DB_PASS, host = DB_HOST, port = DB_PORT)
    cursor = cnx.cursor()

def disconnectMysql():
    global cnx
    global cursor
    if (None != cursor):
        cursor.close()
        cursor = None
    if (None != cnx):
        cnx.commit()
        cnx.close()
        cnx = None

def createDatabase():
    global cursor
    try:
        logger.info("Creating database {}: ".format(DB_NAME))
        cursor.execute("CREATE DATABASE {} DEFAULT CHARACTER SET 'utf8'".format(DB_NAME))
    except mysql.connector.Error as err:
        logger.critical("Failed creating database {}: '{}'".format(DB_NAME, err))
        exit(1)

def createTables():
    global cursor
    for name, ddl in TABLES.iteritems():
        try:
            logger.info("Creating table {}: ".format(name))
            cursor.execute(ddl)
        except mysql.connector.Error as err:
            if err.errno == errorcode.ER_TABLE_EXISTS_ERROR:
                logger.info("Table {} already exists.".format(name))
            else:
                logger.critical("Failed creating table {}: '{}'".format(name, err))
                exit(1)

def updateTable(sql, expectedError = None, info = ""):
    try:
        logger.info(sql)
        cursor.execute(sql)
    except mysql.connector.Error as err:
        if err.errno == expectedError:
            logger.info("    " + info)
        else:
            logger.critical("Failed '{}': '{}'".format(sql, err))
            exit(1)

def updateDatabase_0_2():
    global cursor
    logger.info("Updating database {} from format 0.1.x to format 0.2.x: ".format(DB_NAME))

    def getMissingAzimuthEntries():
        return queryTable("`id`, `lat_stroke`, `lon_stroke`, `lat_station`, `lon_station`", "`strokes`", "`azimuth` = 0.0 LIMIT 1000")

    def calcDistanceAndAzimuth(stroke):
        lat1 = float(stroke["lat_station"])
        lon1 = float(stroke["lon_station"])
        lat2 = float(stroke["lat_stroke"])
        lon2 = float(stroke["lon_stroke"])
        geodesic = Geodesic.WGS84.Inverse(lat1, lon1, lat2, lon2)
        distance = geodesic["s12"] # distance between station and stroke (meters)
        distance = round((distance / 1000), 1) # distance between station and stroke (km)
        azimuth = round(geodesic["azi1"], 1) # azimuth at station pointing to the stroke (degrees) in the range [-180, 180]
        if (azimuth <= 0): # exclude 0.0 as 0.0 is the marker for not yet calculated
            azimuth += 360 # azimuth at station pointing to the stroke (degrees) in the range ]0, 360]
        return distance, azimuth

    def updateDistanceAndAzimuth(stroke):
        global cursor
        done = False
        sql = "UPDATE `strokes` SET `distance` = {}, `azimuth` = {} WHERE id = {}".format(stroke["distance"], stroke["azimuth"], stroke["id"])
        try:
            cursor.execute(sql)
            done = True
        except mysql.connector.Error as err:
            logger.warning(err)
        return done

    def updateDatabaseStructure():
        sql = "ALTER TABLE `strokes` DROP `alt_stroke`"
        updateTable(sql, errorcode.ER_CANT_DROP_FIELD_OR_KEY, "column `strokes`.`alt_stroke` is already dropped")

        sql = "ALTER TABLE `strokes` DROP `alt_station`"
        updateTable(sql, errorcode.ER_CANT_DROP_FIELD_OR_KEY, "column `strokes`.`alt_station` is already dropped")

        sql = "ALTER TABLE `strokes` ADD `azimuth` decimal(4,1) NOT NULL AFTER `distance`"
        updateTable(sql, errorcode.ER_DUP_FIELDNAME, "column `strokes`.`azimuth` is already inserted")
    def updateDatabaseContent():
        strokes = getMissingAzimuthEntries()
        updated = 0
        failed = 0
        while strokes:
            for stroke in strokes:
                stroke["distance"], stroke["azimuth"] = calcDistanceAndAzimuth(stroke)
                if updateDistanceAndAzimuth(stroke):
                    updated += 1
                else:
                    failed += 1
                    logger.error("failed to update {}".format(stroke))
            cnx.commit()
            logger.info("updated = {}, failed = {}".format(updated, failed))
            strokes = getMissingAzimuthEntries()
        logger.info("updated = {}, failed = {}".format(updated, failed))

    updateDatabaseStructure()
    updateDatabaseContent()

def updateDatabase_0_3():
    global cursor
    logger.info("Updating database {} from format 0.2.x to format 0.3.x: ".format(DB_NAME))

    def addIndexToTable(table, index, indexNames):
        sql = "ALTER TABLE `{}` ADD INDEX(`{}`);".format(table, index)
        if index not in indexNames:
            updateTable(sql)
        else:
            logger.info(sql)
            logger.info("    index `{}` is already present".format(index));

    def updateDatabaseStructure_part1():
        sql = "ALTER TABLE `signals` ADD `channels` tinyint(4) UNSIGNED NOT NULL AFTER `time_ns`";
        updateTable(sql, errorcode.ER_DUP_FIELDNAME, "column `signals`.`channels` is already inserted")

        sql = "ALTER TABLE `strokes` ADD `channels` tinyint(4) UNSIGNED NOT NULL AFTER `azimuth`";
        updateTable(sql, errorcode.ER_DUP_FIELDNAME, "column `strokes`.`channels` is already inserted")

        sql = "SHOW INDEX FROM `strokes`"
        result = query(sql)
        indexNames = []
        for i in result:
            indexNames.append(i["Key_name"])

        addIndexToTable("signals", "channels", indexNames)
        addIndexToTable("strokes", "azimuth", indexNames)
        addIndexToTable("strokes", "channels", indexNames)
        addIndexToTable("strokes", "distance", indexNames)
        addIndexToTable("strokes", "time_ns_stroke", indexNames)

        sql = "CREATE TABLE `historicalData` ( `id` INT NOT NULL AUTO_INCREMENT, `entity` TINYTEXT NOT NULL, `attribute` TINYTEXT NOT NULL, `value` TEXT NOT NULL, PRIMARY KEY (`id`)) ENGINE=InnoDB DEFAULT CHARSET=utf8;"
        updateTable(sql, errorcode.ER_TABLE_EXISTS_ERROR, "table `historicalData` is already created")

    def getTableColumnNames(table):
        sql = "SELECT * FROM `{}` LIMIT 1".format(table)
        result = query(sql)
        columnNames = []
        if result:
            for i in result[0]:
                columnNames.append(i)
        return columnNames

    def updateDatabaseContent():
        """converts the columns `channel0`, `channel1`, `channel2`, `channel3` to column `channels`"""
        columnNames = getTableColumnNames("signals")
        for c in range(4):
            if ("channel{}".format(c) in columnNames):
                sql = "UPDATE `signals` SET `channels`= `channels` | (`channel{c}` << {c}) WHERE `channel{c}` IS NOT NULL".format(c = c)
                updateTable(sql)

    def updateDatabaseStructure_part2():
        columnNames = getTableColumnNames("signals")
        for c in range(4):
            errors = 0
            if ("channel{}".format(c) in columnNames):
                select = "count(*) as errors"
                table = "`signals`"
                where = "((`channels` & (1 << {c})) != (`channel{c}` << {c})) AND `channel{c}` IS NOT NULL".format(c = c)
                result = queryTable(select, table, where)
                errors = result[0]["errors"]
            if (0 == errors):
                sql = "ALTER TABLE `signals` DROP `channel{c}`".format(c = c)
                updateTable(sql, errorcode.ER_CANT_DROP_FIELD_OR_KEY, "column `signals`.`channel{c}` is already dropped".format(c = c))
        sql = "UPDATE `strokes`, `signals` SET strokes.channels = signals.channels WHERE strokes.signal_id = signals.id"
        updateTable(sql)

    updateDatabaseStructure_part1()
    updateDatabaseContent()
    updateDatabaseStructure_part2()

def updateDatabase():
    global DB_NAME
    logger.info("-> updateDatabase()")
    connectMysql()
    connectDatabase()
    updateDatabase_0_2()
    updateDatabase_0_3()
    disconnectMysql()
    logger.info("<- updateDatabase()")

def connectDatabase():
    global cnx
    try:
        cnx.database = DB_NAME
    except mysql.connector.Error as err:
        if err.errno == errorcode.ER_BAD_DB_ERROR:
            createDatabase()
            cnx.database = DB_NAME
            createTables()
        else:
            logger.critical("Failed to connect to database {}: '{}'".format(DB_NAME, err))
            exit(1)

def insertSignals(signals):
    global cursor
    insert = "`time_ns`, `channels`"
    values = ""
    numDone = 0
    for signal in signals:
        if ("time_ns" in signal) and ("channels" in signal) and (0 != (signal["channels"] & MAX_STORABLE_NUMBER_OF_CHANNELS_MASK)):
            values += "({}, {}), ".format(signal["time_ns"], signal["channels"])
            numDone += 1
        else:
            logger.error("insertSignals({}): 'time_ns' or 'channels' missing or no channel triggered".format(signal))
    sql = "INSERT IGNORE INTO `signals` ({}) VALUES {}".format(insert, values[0:-2]) # -2 means drop ", "
    try:
        cursor.execute(sql)
        cnx.commit()
    except mysql.connector.Error as err:
        logger.warning(err)
        numDone = 0
    return numDone

def insertStrokes(strokes):
    global cursor
    keys = ["time_ns_stroke", "lat_stroke", "lon_stroke", "delta_time_ns_station", "lat_station", "lon_station", "distance", "azimuth"]
    insert = ""
    for key in keys:
        insert += "`{}`, ".format(key)
    values = ""
    numDone = 0
    for stroke in strokes:
        tmp = ""
        failed = False
        for key in keys:
            if (key in stroke):
                tmp += "{}, ".format(stroke[key])
            else:
                logger.error("key '{}' not found in stroke".format(key))
                failed = True
        if (not failed):
            values += "({}), ".format(tmp[0:-2]) # -2 means drop ", "
            numDone += 1
    sql = "INSERT IGNORE INTO `strokes` ({}) VALUES {}".format(insert[0:-2], values[0:-2]) # -2 means drop ", "
    try:
        cursor.execute(sql)
        cnx.commit()
    except mysql.connector.Error as err:
        logger.warning(err)
        numDone = 0
    return numDone

def updateStrokeSignal(link):
    global cursor
    done = False
    sql = "UPDATE `strokes` SET `signal_id` = {}, `channels` = {} WHERE `id` = {}".format(link["signal_id"], link["channels"], link["stroke_id"])
    try:
        cursor.execute(sql)
        cnx.commit()
        done = True
    except mysql.connector.Error as err:
        logger.warning(err)
    return done

def query(sql):
    global cursor
    result = []
    try:
        cursor.execute(sql)
        columns = cursor.description 
        result = [{columns[index][0]:column for index, column in enumerate(value)} for value in cursor.fetchall()]
        for i in range(len(result)):
            for key in result[i]:
                if (type(Decimal(0)) == type(result[i][key])):
                    result[i][key] = float(result[i][key])
    except mysql.connector.Error as err:
        logger.warning(err)
    return result

def queryTable(select, tables, where = None, limit = None):
    sql = "SELECT {} FROM {}".format(select, tables)
    if where:
        sql += " WHERE {}".format(where)
    if limit:
        sql += " LIMIT {}".format(limit)
    return query(sql)

def querySignals(select = "*", where = 1):
    return queryTable(select, "`signals`", where)

def queryStrokes(select = "*", where = 1):
    return queryTable(select, "`strokes`", where)

def correlateSignalsStrokesByTime(timeCorrection):
    select = "signals.channels as channels, signals.id as signal_id, strokes.id as stroke_id"
    tables = "`signals`, `strokes`"
    where = "strokes.time_ns_stroke + strokes.delta_time_ns_station + {} = signals.time_ns AND strokes.signal_id IS NULL".format(timeCorrection)
    return queryTable(select, tables, where)

def correlateSignalsStrokes():
    # by manual correlation I found the data on the server being systematically 1 sec = 1000000000 ns off; in few cases it was 1 ns less: 999999999 ns
    links = correlateSignalsStrokesByTime(-1000000000)
    links += correlateSignalsStrokesByTime(-999999999)
    linkCnt = 0
    linkUpdateFailed = 0
    for link in links:
        if not updateStrokeSignal(link):
            linkUpdateFailed += 1
        linkCnt += 1
    logger.setLevel(logging.DEBUG)
    if (linkUpdateFailed):
        logger.info("correlating signals and strokes: {} links identified, update failed for {}".format(linkCnt, linkUpdateFailed))
    else:
        logger.info("correlating signals and strokes: {} links identified".format(linkCnt))
    logger.setLevel(LOGLEVEL)

def testInsertSignals():
    logger.info("-> testInsertSignals()")
    signals = []

    logger.info("# test error: channels missing")
    signal = {'time_ns': 12345678901}
    signals.append(signal) # collect data

    logger.info("# test error: time_ns missing")
    signal = {'channels': 0x0F}
    signals.append(signal) # collect data

    logger.info("# test error: both present but all channels unused")
    signal = {'time_ns': 12345678902, 'channels': 0x00}
    signals.append(signal) # collect data

    logger.info("# test error: wrong number of channels")
    signal = {'time_ns': 12345678903, 'channels': 0x100}
    signals.append(signal) # collect data

    logger.info("# test valid: 1 channel")
    signal = {'time_ns': 1483871710203059998L + 7133192 - 1000000000, 'channels': 0x01}
    signals.append(signal) # collect data

    logger.info("# test valid: 8 channels")
    signal = {'time_ns': 1483871710203059999L + 7133193 - 999999999, 'channels': 0xFF}
    signals.append(signal) # collect data

    logger.info("# test valid: 4 channels")
    signal = {'time_ns': 12345678906, 'channels': 0x0F}
    signals.append(signal) # collect data

    logger.info("# test silently ignore repeated key 'time_ns'")
    signal = {'time_ns': 12345678906, 'channels': 0x0F}
    signals.append(signal) # collect data

    numDone = insertSignals(signals) # insert data into the database
    print("insertSignals() has done {}".format(numDone))
    logger.info("<- testInsertSignals()")

def testQuerySignals():
    logger.info("-> testQuerySignals()")
    logger.info("# test default")
    result = querySignals()
    logger.debug(json.dumps(result, indent=4))

    logger.info("# test with select and where clause")
    channel = 0
    result = querySignals(select = "`id`, `time_ns`", where = "`channels` & (1 << {}) = (1 << {})".format(channel, channel))
    logger.debug(json.dumps(result, indent=4))
    logger.info("<- testQuerySignals()")

def testInsertStrokes():
    logger.info("-> testInsertStrokes()")
    strokes = []

    logger.info("# test error: key 'time_ns_stroke' missing")
    stroke = {'lat_stroke': 37.980775, 'lon_stroke': 24.618726, 'delta_time_ns_station': 7133192, 'lat_station': 50.348629, 'lon_station': 4.24392, 'distance': 2119.2, 'azimuth': 0.0}
    strokes.append(stroke) # collect data

    logger.info("# test error: key 'lat_stroke' missing")
    stroke = {'time_ns_stroke': 1483871710203059998L, 'lon_stroke': 24.618726, 'delta_time_ns_station': 7133192, 'lat_station': 50.348629, 'lon_station': 4.24392, 'distance': 2119.2, 'azimuth': 0.0}
    strokes.append(stroke) # collect data

    logger.info("# test error: key 'lon_stroke' missing")
    stroke = {'time_ns_stroke': 1483871710203059998L, 'lat_stroke': 37.980775, 'delta_time_ns_station': 7133192, 'lat_station': 50.348629, 'lon_station': 4.24392, 'distance': 2119.2, 'azimuth': 0.0}
    strokes.append(stroke) # collect data

    logger.info("# test error: key 'delta_time_ns_station' missing")
    stroke = {'time_ns_stroke': 1483871710203059998L, 'lat_stroke': 37.980775, 'lon_stroke': 24.618726, 'lat_station': 50.348629, 'lon_station': 4.24392, 'distance': 2119.2, 'azimuth': 0.0}
    strokes.append(stroke) # collect data

    logger.info("# test error: key 'lat_station' missing")
    stroke = {'time_ns_stroke': 1483871710203059998L, 'lat_stroke': 37.980775, 'lon_stroke': 24.618726, 'delta_time_ns_station': 7133192, 'lon_station': 4.24392, 'distance': 2119.2, 'azimuth': 0.0}
    strokes.append(stroke) # collect data

    logger.info("# test error: key 'lon_station' missing")
    stroke = {'time_ns_stroke': 1483871710203059998L, 'lat_stroke': 37.980775, 'lon_stroke': 24.618726, 'delta_time_ns_station': 7133192, 'lat_station': 50.348629, 'distance': 2119.2, 'azimuth': 0.0}
    strokes.append(stroke) # collect data

    logger.info("# test error: key 'distance' missing")
    stroke = {'time_ns_stroke': 1483871710203059998L, 'lat_stroke': 37.980775, 'lon_stroke': 24.618726, 'delta_time_ns_station': 7133192, 'lat_station': 50.348629, 'lon_station': 4.24392, 'azimuth': 0.0}
    strokes.append(stroke) # collect data

    logger.info("# test error: key 'azimuth' missing")
    stroke = {'time_ns_stroke': 1483871710203059998L, 'lat_stroke': 37.980775, 'lon_stroke': 24.618726, 'delta_time_ns_station': 7133192, 'lat_station': 50.348629, 'lon_station': 4.24392, 'distance': 2119.2}
    strokes.append(stroke) # collect data

    logger.info("# test valid 1")
    stroke = {'time_ns_stroke': 1483871710203059998L, 'lat_stroke': 37.980775, 'lon_stroke': 24.618726, 'delta_time_ns_station': 7133192, 'lat_station': 50.348629, 'lon_station': 4.24392, 'distance': 2119.2, 'azimuth': 0.0}
    strokes.append(stroke) # collect data

    logger.info("# test valid 2")
    stroke = {'time_ns_stroke': 1483871710203059999L, 'lat_stroke': 37.980776, 'lon_stroke': 24.618727, 'delta_time_ns_station': 7133193, 'lat_station': 50.348630, 'lon_station': 4.24393, 'distance': 2119.3, 'azimuth': 0.0}
    strokes.append(stroke) # collect data

    logger.info("# test silently ignore repeated key 'time_ns_stroke' + 'delta_time_ns_station'")
    stroke = {'time_ns_stroke': 1483871710203059998L, 'lat_stroke': 37.980775, 'lon_stroke': 24.618726, 'delta_time_ns_station': 7133192, 'lat_station': 50.348629, 'lon_station': 4.24392, 'distance': 2119.2, 'azimuth': 0.0}
    strokes.append(stroke) # collect data

    numDone = insertStrokes(strokes) # insert data into the database
    print("insertStrokes() has done {}".format(numDone))
    logger.info("<- testInsertStrokes()")

def testQueryStrokes():
    logger.info("-> testQueryStrokes()")
    logger.info("# test default")
    result = queryStrokes()
    logger.debug(json.dumps(result, indent=4))

    logger.info("# test with select and where clause, no data")
    result = queryStrokes(select = "`id`, `distance`, `azimuth`", where = "`time_ns_stroke` = 0")
    logger.debug(json.dumps(result, indent=4))

    logger.info("# test with select and where clause, data")
    result = queryStrokes(select = "`id`, `distance`, `azimuth`", where = "`time_ns_stroke` = 1483871710203059998")
    logger.debug(json.dumps(result, indent=4))
    logger.info("<- testQueryStrokes()")

def test():
    global DB_NAME
    DB_NAME = 'testBoSignals'
    logger.info("-> test()")

    # test updating the databse (first time the structure is updated)
    updateDatabase()

    # test connect
    connectMysql()
    connectDatabase()

    # tests for insert and query
    testInsertSignals()
    testQuerySignals()
    testInsertStrokes()
    correlateSignalsStrokes()
    testQueryStrokes()

    # test disconnect
    disconnectMysql()

    # test updating the databse (second time the structure update shall not fail and the content is updated)
    updateDatabase()

    logger.info("<- test()")

def configLogger():
    global logger
    if __name__ == "__main__":
        format = "%(levelname)s: %(message)s"
        logging.basicConfig(format = format)
    logger = logging.getLogger(__name__)
    if __name__ == "__main__":
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(LOGLEVEL)

def main():
    configLogger()
    if (2 == len(sys.argv)) and ("--updateDatabase" == sys.argv[1]):
        updateDatabase()
    else:
        test() # optionally enable test in order to run some quick tests on a pure test database

if __name__ == "__main__":
    main()
