#!/usr/bin/python
# the thrading code is based on https://www.tutorialspoint.com/python/python_multithreading.htm

from __future__ import print_function
from boConfig import FTP_LOCAL_PATH, IMAGES, BOOK_KEEPING_PATH, NUM_IMAGE_THREADS
import subprocess
import sys
import os
import json
import time
import Queue
import threading

TIME_STEP = 10 * 60 # the Blitzortung server updates the signal and stroke data every 10 minutes
FOERVER = 10 * 52 * 7 * 24 * 60 * 60 # for the simplicity of the calculation 10 years is the equivalent of 'forever'

bookKeepingFile = "boGenerateImages.json"

# thread management
workQueue = None
workQueueLock = None
resultQueue = None
resultQueueLock = None
exitFlag = False

class myThread(threading.Thread):
    def __init__(self, threadID, name, workQueue, resultQueue):
        threading.Thread.__init__(self)
        self.threadID = threadID
        self.name = name
        self.workQueue = workQueue
        self.resultQueue = resultQueue

    def run(self):
        #print("Starting " + self.name)
        process_data(self.name, self.workQueue, self.resultQueue)
        #print("Exiting " + self.name)

def process_data(threadName, workQueue, resultQueue):
    global workQueueLock
    global resultQueueLock
    global exitFlag

    while not exitFlag:
        workQueueLock.acquire()
        if not workQueue.empty():
            params = workQueue.get()
            workQueueLock.release()
            #print("{} start '{}'".format(threadName, params))
            imageData = generateImage(params)
            #print("{} done  '{}'".format(threadName, params))
            resultQueueLock.acquire()
            resultQueue.put([params, imageData])
            resultQueueLock.release()
        else:
            workQueueLock.release()
            time.sleep(0.01)

def calcEndOfGenerationTime():
    endTime = int(time.time() / TIME_STEP) * TIME_STEP # now rounded down by TIME_STEP example: now is 08:13:54 then endTime is 08:10:00
    endTime += TIME_STEP # end of the script execution shall be 08:20:00
    endTime -= 15 # keep 15 seconds as reserve for the FTP transfer, end of the image generation shall be at 08:19:45
    return endTime

def parseTime(t = None):
    factor = {
        'y': 52 * 7 * 24 * 60 * 60, # one (simplified) year of 52 weeks
        'w': 7 * 24 * 60 * 60, # one week
        'd': 24 * 60 * 60, # one day
        'h': 60 * 60, # one hour
        'm': 60, # one minute
        's': 1 # one second
        }
    if (None == t):
        seconds = FOERVER
    else:
        val = int(t[0:-1])
        unit = t[-1]
        if unit in factor:
            seconds = val * factor[unit]
        else:
            seconds = FOERVER
    return seconds

def writeDict(file, dict):
    try:
        json.dump(dict, open(file, 'w'), indent=4, sort_keys=True)
    except:
        print("ERROR: could not write file '{}'".format(file))
        e = sys.exc_info()[0]
        print( "ERROR: {}".format(e))

def readDict(file):
    dict = {}
    try:
        dict = json.load(open(file, 'r'))
    except:
        print("ERROR: could not read file '{}'".format(file))
        e = sys.exc_info()[0]
        print( "ERROR: {}".format(e))
    return dict

def cleanupImageInformation(imageInformation, paramsList):
    toBeRemoved = []
    for key in imageInformation:
        if (key not in paramsList):
            toBeRemoved.append(key)
    for key in toBeRemoved:
        del(imageInformation[key])
    return imageInformation

def calcImageTimeCoverage(imageInformation, paramsList):
    for params in paramsList:
        seconds = FOERVER
        for param in params.split():
            if ("-t" == param[0:2]):
                seconds = parseTime(param[2:])
                break
            elif ("-i5" == param): # signals(channel), strokes(channel)/signals(channel), 24h, channel X
                seconds = 24 * 60 * 60 # h/d * m/h * s/m => s/d
                break
            elif ("-i8" == param): # strokes(channel) and locating ratio 14d
                seconds = 14 * 24 * 60 * 60 # d/fortnight * h/d * m/h * s/m => s/fortnight
                break
            elif ("-i9" == param): # strokes(channel) and locating ratio 24h
                seconds = 24 * 60 * 60 # h/d * m/h * s/m => h/d
                break
            elif ("-n1000" == param): # -i10, map channel strokes on MAP_WORLD, all channels, 1000 most distant
                seconds = 24 * 60 * 60 # h/d * m/h * s/m => h/d
                break
            elif ("-i12" == param): # strokes(channel) and locating ratio 1y
                seconds = 365 * 30 * 24 * 60 * 60 # d/y * d/M * h/d * m/h * s/m => s/y
                break
        if params not in imageInformation:
            imageInformation[params] = {}
        imageInformation[params]['timeCoverage'] = seconds
    return imageInformation

def calcPriority(priority, timeCoverage):
    if (None == priority):
        priority = 0
        while(timeCoverage > 10):
            timeCoverage /= 10
            priority += 1
    else:
        priority -= 1
    return priority # a small number is urgent, a high number can wait

def calcSortedImageInformation(imageInformation, paramsList):
    for params in paramsList:
        if (params in imageInformation):
            if ('priority' not in imageInformation[params]):
                imageInformation[params]['priority'] = None
            imageInformation[params]['priority'] = calcPriority(imageInformation[params]['priority'], imageInformation[params]['timeCoverage'])
            if ('executionTime' not in imageInformation[params]):
                imageInformation[params]['executionTime'] = 60 # no measured time available, assume 60 seconds required to generate the image
        else:
            print("ERROR: '{}' not found in imageInformation".format(params))
    sortedImageInformation = sorted(imageInformation.iteritems(), key=lambda(x, y): y['priority'])
    return sortedImageInformation

def pickNextImageForGeneration(sortedImageInformation, remainingTime, idleThreads, returnMinimumOne):
    params = None
    
    if len(sortedImageInformation):
        remainingThreadTime = []
        for t in range(idleThreads):
            remainingThreadTime.append(remainingTime)

        imageGenerationPossible = 0
        for i in range(0, len(sortedImageInformation)):
            threadFound = False
            for t in range(idleThreads):
                if (sortedImageInformation[i][1]['executionTime'] < remainingThreadTime[t]):
                    remainingThreadTime[t] -= sortedImageInformation[i][1]['executionTime']
                    threadFound = True
                    imageGenerationPossible += 1
                    break
            if not threadFound:
                break

        if (not imageGenerationPossible) and (returnMinimumOne):
            imageGenerationPossible = 1

        if imageGenerationPossible:
            # pick the longest running diagram first
            maxExecutionTime = 0
            for i in range(imageGenerationPossible):
                if maxExecutionTime < sortedImageInformation[i][1]['executionTime']:
                    maxExecutionTime = sortedImageInformation[i][1]['executionTime']
                    maxExecutionTimeIndex = i

            # get and return the one with the maximum execution time
            params = sortedImageInformation[maxExecutionTimeIndex][0]
            #print("remainingThreadTime = {}, maxExecutionTime = {}, imageGenerationPossible = {}".format(remainingThreadTime, maxExecutionTime, imageGenerationPossible))
            #print("sortedImageInformation[{}] = {}".format(maxExecutionTimeIndex, sortedImageInformation[maxExecutionTimeIndex]))
            sortedImageInformation.remove(sortedImageInformation[maxExecutionTimeIndex])
    return sortedImageInformation, params

def generateImage(params):
    imageData = None
    tStart = time.time()
    cmd = "./boPng.php {}".format(params)
    sp = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = sp.communicate()
    ext = None
    if ("" == err):
        if ("PNG" == out[1:4]):
            ext = "png"
        if ext:
            fileName = "IMG{}.{}".format("".join(i for i in params.split()), ext)
            filePath = "{}/{}".format(FTP_LOCAL_PATH, fileName)
            with open(filePath, "wb") as f:
                f.write(out)
            executionTime = time.time() - tStart
            imageData = {}
            imageData['fileName'] = fileName # required for the HTML generation
            imageData['filePath'] = filePath # required for the HTML generation
            imageData['executionTime'] = executionTime # required as information for the next pass
            imageData['priority'] = None # the generated image has to get a new priority in the next pass
            imageData['generationTime'] = time.time() # information is just informational
            print("{} > {}, {:.1f} sec".format(cmd, fileName, executionTime))
        else:
            print("ERROR: '{}' failed".format(params))
    else:
        print("ERROR: '{}'".format(err))
    return imageData

def generateImages(sortedImageInformation, imageInformation):
    global workQueue
    global workQueueLock
    global resultQueue
    global resultQueueLock
    global exitFlag

    workQueue = Queue.Queue(NUM_IMAGE_THREADS)
    workQueueLock = threading.Lock()
    resultQueue = Queue.Queue(len(IMAGES))
    resultQueueLock = threading.Lock()
    
    threadList = []
    for i in range(NUM_IMAGE_THREADS):
        threadList.append("Thread-{}".format(i + 1))

    # Create new threads
    threads = []
    threadID = 1
    for tName in threadList:
        thread = myThread(threadID, tName, workQueue, resultQueue)
        thread.start()
        threads.append(thread)
        threadID += 1

    # Fill the queue
    startedImageCount = 0
    endTime = calcEndOfGenerationTime()
    while len(sortedImageInformation):
        threadsInUse = startedImageCount - resultQueue.qsize()
        while (NUM_IMAGE_THREADS == threadsInUse):
            time.sleep(0.1)
            threadsInUse = startedImageCount - resultQueue.qsize()

        remainingTime = endTime - int(time.time())
        idleThreads = NUM_IMAGE_THREADS - threadsInUse
        sortedImageInformation, params = pickNextImageForGeneration(sortedImageInformation, remainingTime, idleThreads, 0 == startedImageCount)
        if (params):
            workQueueLock.acquire()
            workQueue.put(params)
            #print("Queuing '{}'".format(params))
            workQueueLock.release()
            startedImageCount += 1
        else:
            break

    # Wait for queue to empty
    while not workQueue.empty():
        time.sleep(0.1)

    # Notify threads it's time to exit
    exitFlag = True

    # Wait for all threads to complete
    for t in threads:
        t.join()

    # Collect the results
    generatedImageCount = 0
    while not resultQueue.empty():
        resultQueueLock.acquire()
        result = resultQueue.get()
        #print(result)
        params = result[0]
        imageData = result[1]
        if (imageData):
            # success
            imageData["timeCoverage"] = imageInformation[params]["timeCoverage"]
            imageInformation[params] = imageData
            generatedImageCount += 1
        else:
            # failed
            pass
        resultQueueLock.release()

    #print("Exiting Main Thread")

    return imageInformation, generatedImageCount

def generateHtml(imageInformation, paramsList):
    availableImageCount = 0
    html = """<html>
    <head>
    </head>
    <body>
"""
    for params in paramsList:
        if ('fileName' in imageInformation[params]) and ('filePath' in imageInformation[params]):
            if os.path.isfile(imageInformation[params]['filePath']):
                html += "        <img src='{}' />\n".format(imageInformation[params]['fileName'])
                availableImageCount += 1
    html += """    </body>
</html>"""
    file = "{}/test.html".format(FTP_LOCAL_PATH)
    with open(file, "w") as f:
        f.write(html)
    print("{} > {}".format(sys.argv[0], file))
    return availableImageCount

def getBookKeepingPath():
    global bookKeepingFile
    if ("" == BOOK_KEEPING_PATH):
        return
    if (os.path.exists(BOOK_KEEPING_PATH) and os.path.isdir(BOOK_KEEPING_PATH)):
        if ('/' == BOOK_KEEPING_PATH[-1]):
            bookKeepingFile = BOOK_KEEPING_PATH + bookKeepingFile
        else:
            bookKeepingFile = BOOK_KEEPING_PATH + '/' + bookKeepingFile
    else:
        print("ERROR: BOOK_KEEPING_PATH = '{}' does not exist".format(BOOK_KEEPING_PATH))
        sys.exit(1)

def main():
    global bookKeepingFile
    paramsList = IMAGES
    if len(sys.argv) > 1:
        paramsList = [''.join("{} ".format(i) for i in sys.argv[1:])]

    getBookKeepingPath()
    imageInformation = readDict(bookKeepingFile)
    imageInformation = cleanupImageInformation(imageInformation, paramsList)
    imageInformation = calcImageTimeCoverage(imageInformation, paramsList)
    sortedImageInformation = calcSortedImageInformation(imageInformation, paramsList)
    imageInformation, generatedImageCount = generateImages(sortedImageInformation, imageInformation)
    if len(sys.argv) == 1:
        writeDict(bookKeepingFile, imageInformation)

    availableImageCount = generateHtml(imageInformation, paramsList)
    print("{} images generated, {} images available".format(generatedImageCount, availableImageCount))

if __name__ == "__main__":
    main()
