#!/usr/bin/python
from __future__ import print_function
from boConfig import BO_USER, BO_PASS, BO_STATION, BO_REGION, BOOK_KEEPING_PATH, LOGGING_PATH, LOGLEVEL
import time
import calendar
import subprocess
import json
import boDb
import sys
import os
import logging
from geographiclib.geodesic import Geodesic

desiredExecutionTime = 4 * 60 # note: cron starts this job every 10 minutes, thus some reserve should be there
secStep = 10 * 60 # 10 minutes in seconds
bookKeepingFile = "boGetStrokesFromBlitzortung.txt"
logFile = "boGetStrokesFromBlitzortung.log"

url = "http://{}:{}@data.blitzortung.org/Data_{}/Protected/Strokes/%Y/%m/%d/%H/%M.json.gz"

new = 0
done = 0
dropped = 0

def configLogger():
    global logger
    global logFile
    format = "%(asctime)-15s %(levelname)s:%(name)s: %(message)s"

    path = LOGGING_PATH
    if ("" == path):
        path = "./"
    if (os.path.exists(path) and os.path.isdir(path)):
        if ('/' == path[-1]):
            filename = path + logFile
        else:
            filename = path + '/' + logFile
    else:
        print("ERROR: LOGGING_PATH = '{}' does not exist".format(LOGGING_PATH))
        sys.exit(1)

    logging.basicConfig(filename = filename, format = format)
    logger = logging.getLogger(__name__)
    logger.setLevel(LOGLEVEL)
    logger.setLevel(logging.INFO) # overrule LOGLEVEL
    boDb.configLogger()

def calcRequiredDownloads():
    """1 download to be done every 10 minutes from 3 servers
    server may not be reachable => retrials may be required
    server may also not have data => no endless retries
    file for bookkeeping
    first line: start and end time/date in human readable format
    second line: end time/date in human readable format
    following lines: downloads to be done from start to end (inclduing borders))
    succeeded lines are deleted
    missing lines for newer dates are to be added
    """
    global secStep
    global bookKeepingFile
    global new

    fileDateTimeFormat = "%Y.%m.%d %H:%M"
    firstLine = True

    urls = []
    secEnd = (int(time.time() / secStep) * secStep) - secStep # multiple of secStep, older than now
    try:
        with open(bookKeepingFile, 'rt') as f:
            for line in f:
                line = line.rstrip()
                if (line):
                    urls.append(line)
                if (firstLine):
                    firstLine = False
                    line = line.split('-')
                    tInitialStart = time.strptime(line[0].strip(), fileDateTimeFormat)
                    secInitialStart = int(calendar.timegm(tInitialStart) / secStep) * secStep
                    tInitialEnd = time.strptime(line[1].strip(), fileDateTimeFormat)
                    secInitialEnd = int(calendar.timegm(tInitialEnd) / secStep) * secStep
                    secStart = secInitialEnd + secStep
    except:
        secInitialStart = secEnd - 3 * 60 * 60 # signals are available for the previous 3 hours, thus initially start with this time range
        secStart = secInitialStart
        urls.append("")
    urls[0] = time.strftime(fileDateTimeFormat, time.gmtime(secInitialStart)) + " - " + time.strftime(fileDateTimeFormat, time.gmtime(secEnd))

    new = 0
    for sec in range(secStart, secEnd + secStep, secStep):
        t = time.gmtime(sec)
        for i in BO_REGION:
            link = url.format(BO_USER, BO_PASS, i)
            link = time.strftime(link, t)
            urls.append(link)
            new += 1
    return urls

def savePendingDownloads(urls):
    global new
    global done
    global dropped
    with open(bookKeepingFile, 'wt') as f:
        for line in urls:
            f.write(line + '\n')
    logger.info("{} downloaded: {} new, {} done, {} dropped, {} pending".format(urls[0], new, done, dropped, len(urls) - 1))

def readStderr(sp, cmd, maxReturnCode = 0):
    err = ""
    if (sp.returncode > maxReturnCode):
        err += "{} returncode: {}; ".format(cmd, sp.returncode)
    line = sp.stderr.readline()
    while line:
        err += "{}: {}; ".format(cmd, line)
        line = sp.stderr.readline()
    return err

def getStrokeStationData(line):
    stroke = {}
    j = {}
    try:
        j = json.loads(line)
    except:
        logger.error("parsing json data failed: '{}'".format(sys.exc_info()[0]))
    if ("sig" in j):
        for i in reversed(range(len(j["sig"]))):
            if (int(BO_STATION) == j["sig"][i]["sta"]):
                stroke["time_ns_stroke"] = j["time"]
                stroke["lat_stroke"] = j["lat"]
                stroke["lon_stroke"] = j["lon"]
                stroke["delta_time_ns_station"] = j["sig"][i]["time"]
                stroke["lat_station"] = j["sig"][i]["lat"]
                stroke["lon_station"] = j["sig"][i]["lon"]
                break
    else:
        logger.error("no 'sig' in the data line '{}'".format(line))
    return stroke

def calcDistanceAndAzimuth(stroke):
    lat1 = float(stroke["lat_station"])
    lon1 = float(stroke["lon_station"])
    lat2 = float(stroke["lat_stroke"])
    lon2 = float(stroke["lon_stroke"])
    geodesic = Geodesic.WGS84.Inverse(lat1, lon1, lat2, lon2)
    distance = geodesic["s12"] # distance between station and stroke (meters)
    distance = round((distance / 1000), 1) # distance between station and stroke (km)
    azimuth = round(geodesic["azi1"], 1) # azimuth at station pointing to the stroke (degrees) in the range [-180, 180]
    if (azimuth <= 0): # exclude 0.0 as 0.0 is the marker for not yet calculated
        azimuth += 360 # azimuth at station pointing to the stroke (degrees) in the range ]0, 360]
    return distance, azimuth

def urlGetGrepParse(url):
    result = False

    cmd1 = "wget -qO- {}".format(url).split()
    sp1 = subprocess.Popen(cmd1, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    cmd2 = "grep \"sta\":{},".format(BO_STATION).split()
    sp2 = subprocess.Popen(cmd2, stdin=sp1.stdout, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    strokes = []
    strokeCnt = 0
    strokeInsertFailed = 0
    line = sp2.stdout.readline()
    while line:
        stroke = getStrokeStationData(line)
        if (stroke):
            stroke["distance"], stroke["azimuth"] = calcDistanceAndAzimuth(stroke)
            strokes.append(stroke)
            strokeCnt += 1
        line = sp2.stdout.readline()
    numDone = boDb.insertStrokes(strokes)
    strokeInsertFailed = strokeCnt - numDone

    sp2.wait() # wait for the process to terminate in order to gte the returncode
    sp1.wait() # wait for the process to terminate in order to gte the returncode
    err = readStderr(sp1, "wget")
    err += readStderr(sp2, "grep", 1) # 0 = match, 1 = no match, 2 = error

    urlForLogging = url.replace("//{}:{}@".format(BO_USER, BO_PASS), "//user:pass@", 1)
    if ("" == err):
        if (strokeInsertFailed):
            logger.info("'{}' contained {} strokes, insert failed for {}".format(urlForLogging, strokeCnt, strokeInsertFailed))
        else:
            logger.info("'{}' contained {} strokes".format(urlForLogging, strokeCnt))
        return True
    else:
        logger.debug("'{}' failed with error '{}'".format(urlForLogging, err.strip()))
        return False

def urlGetUnzipGrepParse(url):
    result = False

    cmd1 = "wget -qO- {}".format(url).split()
    sp1 = subprocess.Popen(cmd1, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    cmd2 = "gunzip -c".split()
    sp2 = subprocess.Popen(cmd2, stdin=sp1.stdout, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    cmd3 = "grep \"sta\":{},".format(BO_STATION).split()
    sp3 = subprocess.Popen(cmd3, stdin=sp2.stdout, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    strokes = []
    strokeCnt = 0
    strokeInsertFailed = 0
    line = sp3.stdout.readline()
    while line:
        stroke = getStrokeStationData(line)
        if (stroke):
            stroke["distance"], stroke["azimuth"] = calcDistanceAndAzimuth(stroke)
            strokes.append(stroke)
            strokeCnt += 1
        line = sp3.stdout.readline()
    numDone = boDb.insertStrokes(strokes)
    strokeInsertFailed = strokeCnt - numDone

    sp3.wait() # wait for the process to terminate in order to gte the returncode
    sp2.wait() # wait for the process to terminate in order to gte the returncode
    sp1.wait() # wait for the process to terminate in order to gte the returncode
    err = readStderr(sp1, "wget")
    err += readStderr(sp2, "gunzip")
    err += readStderr(sp3, "grep", 1) # 0 = match, 1 = no match, 2 = error

    urlForLogging = url.replace("//{}:{}@".format(BO_USER, BO_PASS), "//user:pass@", 1)
    if ("" == err):
        if (strokeInsertFailed):
            logger.info("'{}' contained {} strokes, insert failed for {}".format(urlForLogging, strokeCnt, strokeInsertFailed))
        else:
            logger.info("'{}' contained {} strokes".format(urlForLogging, strokeCnt))
        return True
    else:
        logger.debug("'{}' failed with error '{}'".format(urlForLogging, err))
        return False

def getBookKeepingPath():
    global bookKeepingFile
    if ("" == BOOK_KEEPING_PATH):
        return
    if (os.path.exists(BOOK_KEEPING_PATH) and os.path.isdir(BOOK_KEEPING_PATH)):
        if ('/' == BOOK_KEEPING_PATH[-1]):
            bookKeepingFile = BOOK_KEEPING_PATH + bookKeepingFile
        else:
            bookKeepingFile = BOOK_KEEPING_PATH + '/' + bookKeepingFile
    else:
        logger.error("BOOK_KEEPING_PATH = '{}' does not exist".format(BOOK_KEEPING_PATH))
        sys.exit(1)

def main():
    # normal program starts here
    global done
    global dropped
    global desiredExecutionTime
    done = 0
    dropped = 0
    configLogger()
    tEnd = time.time() + desiredExecutionTime
    getBookKeepingPath()
    boDb.connectMysql()
    boDb.connectDatabase()
    urlsNow = calcRequiredDownloads()
    urlsFuture = []
    urlsFuture.append(urlsNow.pop(0)) # copy start and end date/time
    time.sleep(30) # allow the server to finalize the file - delay due to the fear to miss something not due to knowledge
    while (tEnd > time.time()):
        if (0 == len(urlsNow)):
            break
        url = urlsNow.pop(0).split(',') # try oldest first
        if (1 == len(url)):
            url.append(0) # no error count present, start counting at 0

        # 1st try server 'data' with .gz in order to save bandwidth
        result = urlGetUnzipGrepParse(url[0])

        # 2nd if this failed try server 'data2' with .gz in order to save bandwidth
        if (True != result):
            url[0] = url[0].replace("@data.", "@data2.", 1) # replace server 'data' with 'data2'
            result = urlGetUnzipGrepParse(url[0])

        # 3rd if this failed try sverer 'data2' without .gz
        if (True != result):
            url[0] = url[0][0:-3] # remove '.gz' from the url
            result = urlGetGrepParse(url[0])

        # 4th if this failed, try server 'data' without .gz
        if (True != result):
            url[0] = url[0].replace("@data2.", "@data.", 1) # replace server 'data2' with 'data'
            result = urlGetGrepParse(url[0])

        if (True == result):
            done += 1
        else:
            url[1] = int(url[1]) + 1 # increment the error counter
            urlForLogging = url[0].replace("//{}:{}@".format(BO_USER, BO_PASS), "//user:pass@", 1)
            if (144 <= url[1]): # a day has 24 * 60 = 1440 minutes, best case 1 retry each 10 minutes = 144 retries a day
                # maximum number of retries is reached
                logger.error("'{}' finally failed after {} attempts".format(urlForLogging, url[1]))
                dropped += 1
            else:
                urlsFuture.append("{}.gz,{}".format(url[0], url[1])) # store the incremented error counter
                logger.info("'{}' failed {} times".format(urlForLogging, url[1]))
    for i in range(len(urlsNow)):
        urlsFuture.append("{}".format(urlsNow[i]))
    savePendingDownloads(urlsFuture)
    if done:
        boDb.correlateSignalsStrokes()
    boDb.disconnectMysql()

if __name__ == "__main__":
    main()
